package types

import (
	"errors"
	"fmt"
	"os"
	"path"
	"path/filepath"
	"regexp"
	"strings"
)

var envMatcherRegExp = regexp.MustCompile(`\$[\w\d\_]`)

type (
	AuthType  string
	SmartPath string
)

const (
	AuthTypePasswordless AuthType = "passwordless"
	AuthTypePassword     AuthType = "password"
	AuthTypeKey          AuthType = "key"
)

type Auth struct {
	Type      AuthType  `mapstructure:"type" validate:"required,oneof=passwordless password key"`
	User      string    `mapstructure:"user"`
	Password  string    `mapstructure:"password"`
	Directory SmartPath `mapstructure:"directory"`
	Keyfile   string    `mapstructure:"keyfile"`
}

func (k SmartPath) Resolve(shouldBeDirectory bool) (result string, err error) {
	result = strings.TrimSpace(string(k))
	if strings.HasPrefix(result, "~/") {
		homeDir, err := os.UserHomeDir()
		if err != nil {
			return "", err
		}
		result = path.Join(homeDir, result[2:])
	}
	for _, match := range envMatcherRegExp.FindAllString(string(k), -1) {
		envVar := match[1:]
		if envVar == "" {
			continue
		}
		envVar = os.Getenv(envVar)
		result = strings.ReplaceAll(result, match, envVar)
	}
	result, err = filepath.Abs(result)
	if err != nil {
		return
	}
	stat, err := os.Stat(result)
	if err != nil {
		return
	}
	if !stat.IsDir() && shouldBeDirectory {
		err = errors.New("is not a directory")
	}
	return
}

func (k SmartPath) String() string {
	return string(k)
}

func (a Auth) Validate() error {
	if a.Type == AuthTypePassword && a.Password == "" {
		return fmt.Errorf("password must be provided for authentication type '%s'", AuthTypePassword)
	}
	if a.Type == AuthTypeKey && a.Directory == "" {
		return fmt.Errorf("password must be provided for authentication type '%s'", AuthTypePassword)
	}
	return nil
}
